from __future__ import print_function
from collections import OrderedDict
import copy

import numpy as np
from scipy.optimize import minimize as scipimin

def compute_saxs(q,flags,params):
    """
    Given q, a dict of population flags,
    and a dict of scattering equation parameters,
    compute the saxs spectrum.
    Supported parameters are... TODO: fill in 

    TODO: Document the equation.
    """
    b_flag = flags['bad_data']
    s_flag = flags['diffraction_peaks']
    I = np.zeros(len(q))
    if not b_flag and not s_flag:
        pre_flag = flags['precursor_scattering']
        f_flag = flags['form_factor_scattering']
        I0_floor = params['I0_floor'] 
        I = I0_floor*np.ones(len(q))
        if pre_flag:
            # TODO: This should employ a Guinier-Porod or similar equation
            I0_pre = params['I0_precursor']
            r0_pre = params['r0_precursor']
            I_pre = compute_spherical_normal_saxs(q,r0_pre,0)
            I += I0_pre*I_pre
        if f_flag:
            # TODO: support non-spherical form factors 
            I0_sph = params['I0_sphere']
            r0_sph = params['r0_sphere']
            sigma_sph = params['sigma_sphere']
            I_sph = compute_spherical_normal_saxs(q,r0_sph,sigma_sph)
            I += I0_sph*I_sph
    return I

def profile_spectrum(q,I):
    """
    Profile a saxs spectrum (q,I) 
    by taking several fast numerical metrics 
    from the measured data.

    Returns a dictionary of scalar metrics.
    Dict keys and descriptions:
    'low_q_ratio': fraction of total intensity in the range q<0.4
    'high_q_ratio': fraction of total intensity in the range q>=0.4
    'q_Imax': q value of the maximum intensity.
    'Imax_over_Imean': maximum intensity divided by mean intensity.
    'Imax_over_Ilowq': Maximum intensity divided by mean intensity in the range q<0.4.
    'Imax_over_Ihighq': Maximum intensity divided by mean intensity in the range q>0.4.
    'Imax_sharpness': maximum intensity divided by the mean intensity in the range 0.9*q_Imax<q<1.1*q_Imax.
    'q_bin_edges' : array of q-values to use as upper bin limits for intensity integration.
    'q_bin_strengths' : array of integrated log(I) within the bins specified by q_bin_edges.
    'log_fluctuation': Integrated fluctuation of log(I):
    sum of difference in log(I) between adjacent points,
    taken only where this difference changes sign, divided by the maximum of log(I).
    """ 

    idx_lowq = np.array(q<0.4)
    idx_highq = np.array(q>=0.4)
    # log(I) analysis
    nz = I>0
    q_nz = q[nz]
    I_nz = I[nz]
    logI_nz = np.log(I_nz)
    logI_max = np.max(logI_nz)
    # I analysis
    I_sum = np.sum(I)
    I_mean = np.mean(I)
    I_lowq_mean = np.mean(I[idx_lowq])
    I_highq_mean = np.mean(I[idx_highq])
    idxmax = np.argmax(I)
    idxmin = np.argmin(I)
    Imin = I[idxmin]
    Imax = I[idxmax] 
    q_Imax = q[idxmax]
    Irange = Imax - Imin
    # I_max peak shape analysis
    idx_around_max = ((q > 0.9*q_Imax) & (q < 1.1*q_Imax))
    Imean_around_max = np.mean(I[idx_around_max])
    Imax_sharpness = Imax / Imean_around_max
    # low-q and high-q intensity integration
    dq = q[1:] - q[:-1]
    I_trap = (I[1:]+I[:-1])/2
    I_integral = np.sum(dq*I_trap)
    dq_lowq = q[idx_lowq][1:] - q[idx_lowq][:-1]
    I_trap_lowq = (I[idx_lowq][1:]+I[idx_lowq][:-1])/2
    I_lowq_integral = np.sum(dq_lowq*I_trap_lowq)
    dq_highq = q[idx_highq][1:] - q[idx_highq][:-1]
    I_trap_highq = (I[idx_highq][1:]+I[idx_highq][:-1])/2
    I_highq_integral = np.sum(dq_highq*I_trap_highq)
    # I_max relative intensity analysis
    low_q_ratio = I_lowq_integral / I_integral 
    high_q_ratio = I_highq_integral / I_integral
    Imax_over_Imean = Imax/I_mean
    Imax_over_Ilowq = Imax/I_lowq_mean
    Imax_over_Ihighq = Imax/I_highq_mean

    ### fluctuation analysis
    # array of the difference between neighboring points:
    nn_diff = logI_nz[1:]-logI_nz[:-1]
    # keep indices where the sign of this difference changes.
    # also keep first index
    nn_diff_prod = nn_diff[1:]*nn_diff[:-1]
    idx_keep = np.hstack((np.array([True]),nn_diff_prod<0))
    fluc = np.sum(np.abs(nn_diff[idx_keep]))
    log_fluctuation = fluc/logI_max

    ### bin-integrated log(intensity) analysis
    nbins = 100
    qbinmax = float(1.)
    qbinstep = qbinmax/nbins
    q_bin_edges = np.arange(qbinstep,qbinmax+qbinstep,qbinstep)
    q_bin_strengths = np.zeros(q_bin_edges.shape) 
    binfloor = 0
    for ibin,binmax in zip(range(nbins),q_bin_edges):
        binidx = ((q>=binfloor) & (q<binmax))
        if any(binidx):
            qbin = q[ binidx ]
            Ibin = I[ binidx ]
            dqbin = qbin[1:]-qbin[:-1]
            Ibin = (Ibin[1:]+Ibin[:-1])/2
            q_bin_strengths[ibin] = np.sum(Ibin * dqbin) / I_integral 
        binfloor = binmax
    d = OrderedDict()
    d['q_Imax'] = q_Imax
    d['Imax_over_Imean'] = Imax_over_Imean
    d['Imax_over_Ilowq'] = Imax_over_Ilowq 
    d['Imax_over_Ihighq'] = Imax_over_Ihighq 
    d['Imax_sharpness'] = Imax_sharpness
    d['low_q_ratio'] = low_q_ratio 
    d['high_q_ratio'] = high_q_ratio 
    d['q_bin_edges'] = q_bin_edges
    d['log_fluctuation'] = log_fluctuation
    d['q_bin_strengths'] = q_bin_strengths 
    return d

def parameterize_spectrum(q,I,flags,fixed_params={}):
    """
    Determine a parameterization for a scattering equation,
    beginning with the measured spectrum (q,I) 
    and a dict of population flags. 
    Returns a dict containing the input population flags
    along with initial guesses for the scattering equation parameters
    corresponding to the flagged populations.
    """
    fix_keys = fixed_params.keys()
    d = OrderedDict()
    if flags['bad_data']:
        # stop
        return d 
    else:
        # Get a number for I(q=0)
        if flags['form_factor_scattering']:
            # If form factor scattering, fit 3rd order, use q<0.06. 
            idx_lowq = (q<0.06)
            # Disregard lowest-q values if they are far from the mean, 
            # as these points are likely dominated by experimental error.
            Imean_lowq = np.mean(I[idx_lowq])
            Istd_lowq = np.std(I[idx_lowq])
            idx_good = ((I[idx_lowq] < Imean_lowq+Istd_lowq) & (I[idx_lowq] > Imean_lowq-Istd_lowq))
            I_at_0 = fit_I0(q[idx_lowq][idx_good],I[idx_lowq][idx_good],3)
        elif flags['diffraction_peaks']:
            # If diffraction without form factor scattering, 
            # fit 2nd order, use q<0.06. 
            idx_lowq = (q<0.06)
            I_at_0 = fit_I0(q[idx_lowq],I[idx_lowq],3)
        elif flags['precursor_scattering']:
            # If only precursor scattering, fit third order for entire q-range 
            I_at_0 = fit_I0(q,I,3)
        d['I_at_0'] = I_at_0

    if flags['diffraction_peaks']:
        d['ERROR_MESSAGE']='diffraction peak parameterization not yet supported'
        return d

    if flags['form_factor_scattering']:
        # TODO: insert cases for non-spherical form factors
        #if flags['form_factor_id'] == 'sphere':
            if ('r0_sphere' in fix_keys and 'sigma_sphere' in fix_keys):
                r0_sphere = fixed_params['r0_sphere']
                sigma_sphere = fixed_params['sigma_sphere']
            else:
                # get at least one of r0_sphere or sigma_sphere from spherical_normal_heuristics()
                r0_sphere, sigma_sphere = spherical_normal_heuristics(q,I,I_at_0=I_at_0)
                if 'r0_sphere' in fix_keys:
                    r0_sphere = fixed_params['r0_sphere']
                if 'sigma_sphere' in fix_keys:
                    sigma_sphere = fixed_params['sigma_sphere']
            d['r0_sphere'] = r0_sphere
            d['sigma_sphere'] = sigma_sphere
    
    if flags['precursor_scattering']: 
        if 'r0_precursor' in fix_keys:
            r0_pre = fixed_params['r0_precursor']
        else:
            # TODO: This should employ a Guinier-Porod or similar fit
            # TODO: This approach will have to respect whatever other populations are flagged. 
            r0_pre = precursor_heuristics(q,I,I_at_0=I_at_0)
        d['r0_precursor'] = r0_pre 

    I_floor = np.ones(len(q))
    # include other terms conditionally
    if flags['precursor_scattering']: 
        # TODO: This should be a Guinier-Porod or similar fit
        I_pre = compute_spherical_normal_saxs(q,r0_pre,0)
    if flags['form_factor_scattering']:
        #if flags['form_factor_id'] == 'sphere':
        # TODO: form factors other than sphere
        I_form = compute_spherical_normal_saxs(q,r0_sphere,sigma_sphere)
    #if flags['diffraction_peaks']:
    #   I_structure = .....

    # Run a quick least-squares to get initial guesses for intensity prefactors
    if flags['precursor_scattering'] and flags['form_factor_scattering']:
        x_init = [I_at_0,0.0,0.0]
        x_bounds = [(0.0,None),(0.0,None),(0.0,None)]
        Ifunc = lambda x: x[0]*I_floor + x[1]*I_pre + x[2]*I_form 
    elif flags['precursor_scattering']:
        x_init = [I_at_0,0.0]
        x_bounds = [(0.0,None),(0.0,None)]
        Ifunc = lambda x: x[0]*I_floor + x[1]*I_pre 
    elif flags['form_factor_scattering']:
        x_init = [I_at_0,0.0]
        x_bounds = [(0.0,None),(0.0,None)]
        Ifunc = lambda x: x[0]*I_floor + x[1]*I_form 
    if flags['precursor_scattering'] or flags['form_factor_scattering']:
        I_nz = np.invert((I<=0))
        I_error = lambda x: np.sum( (np.log(Ifunc(x))[I_nz] - np.log(I[I_nz]))**2 )
        res = scipimin(I_error,x_init,bounds=x_bounds,
        constraints=[{'type':'eq','fun':lambda x:np.sum(x)-I_at_0}]) 
        x_res = res.x
        I0_floor = x_res[0]
        if flags['precursor_scattering'] and flags['form_factor_scattering']:
            I0_pre = x_res[1] 
            I0_sphere = x_res[2]
            d['I0_precursor'] = I0_pre 
            d['I0_sphere'] = I0_sphere 
        elif flags['precursor_scattering']:
            I0_pre = x_res[1]
            d['I0_precursor'] = I0_pre 
        elif flags['form_factor_scattering']:
            I0_form = x_res[1]
            d['I0_sphere'] = I0_form 
        d['I0_floor'] = I0_floor 

    #TODO: add parameters for diffraction peaks

    I_guess = compute_saxs(q,flags,d)
    q_I_guess = np.array([q,I_guess]).T
    nz = ((I>0)&(I_guess>0))
    logI_nz = np.log(I[nz])
    logIguess_nz = np.log(I_guess[nz])
    Imean = np.mean(logI_nz)
    Istd = np.std(logI_nz)
    logI_nz_s = (logI_nz - Imean) / Istd
    logIguess_nz_s = (logIguess_nz - Imean) / Istd
    d['R2log_guess'] = compute_Rsquared(logI_nz,logIguess_nz)
    d['chi2log_guess'] = compute_chi2(logI_nz_s,logIguess_nz_s)
    return d

def fit_spectrum(q,I,objfun,flags,params,fit_params,constraints=[]):
    """
    Fit a saxs spectrum (I(q) vs q) to the theoretical spectrum 
    for one or several scattering populations.
    Input objfun (string) specifies objective function to use in optimization.
    Inputs flags (dict) and params (dict) describe flagged scatterer populations
    and initial guesses for the corresponding parameters for the scattering equation.
    Input fit_params (list of strings) indicate the parameters that will be optimized. 
    Input constraints (list of strings) to specify constraints.
    
    Supported objective functions: 
    (1) 'chi2': sum of difference squared across entire q range. 
    (2) 'chi2log': sum of difference of logarithm, squared, across entire q range. 
    (3) 'chi2norm': sum of difference divided by measured value, squared, aross entire q range. 
    (4) 'low_q_chi2': sum of difference squared in only the lowest half of measured q range. 
    (5) 'low_q_chi2log': sum of difference of logarithm, squared, in lowest half of measured q range. 
    (6) 'pearson': pearson correlation between measured and modeled spectra. 
    (7) 'pearson_log': pearson correlation between logarithms of measured and modeled spectra.
    (8) 'low_q_pearson': pearson correlation between measured and modeled spectra. 
    (9) 'low_q_pearson_log': pearson correlation between logarithms of measured and modeled spectra. 

    Supported constraints: 
    (1) 'fix_I0': keeps I(q=0) fixed to the value specified in the input params.

    TODO: document the objective functions, etc.
    """
    pre_flag = flags['precursor_scattering']
    form_flag = flags['form_factor_scattering']
    structure_flag = flags['diffraction_peaks']

    fit_params = copy.deepcopy(fit_params)
    # trim non-flagged populations out of fit_params:
    # fit as few of the requested fit_params as possible
    if not pre_flag:
        if 'I0_precursor' in fit_params:
            fit_params.pop(fit_params.index('I0_precursor')) 
        if 'r0_precursor' in fit_params:
            fit_params.pop(fit_params.index('r0_precursor')) 
    if not form_flag:
        if 'I0_sphere' in fit_params:
            fit_params.pop(fit_params.index('I0_sphere')) 
        if 'r0_sphere' in fit_params:
            fit_params.pop(fit_params.index('r0_sphere')) 
        if 'sigma_sphere' in fit_params:
            fit_params.pop(fit_params.index('sigma_sphere')) 
    #if not structure_flag:
    #    if 'I0_structure' in fit_params:
    #        fit_params.pop(fit_params.index('I0_structure')) 
    #    if 'q0_structure' in fit_params:
    #        fit_params.pop(fit_params.index('q0_structure')) 
    #    if 'sigma_structure' in fit_params:
    #        fit_params.pop(fit_params.index('sigma_structure')) 

    c = []
    if 'fix_I0' in constraints: 
        I_keys = []
        if 'I0_floor' in fit_params:
            I_keys.append('I0_floor')
        if 'I0_precursor' in fit_params:
            I_keys.append('I0_precursor')
        if 'I0_sphere' in fit_params:
            I_keys.append('I0_sphere')
        #if 'I0_structure' in fit_params:
        #    I_keys.append('I0_structure')
        if len(I_keys) == 0:
            # constraint inherently satisfied: do nothing.
            pass
        elif len(I_keys) == 1:
            # satisfy constraint by removing this key from fit_params. 
            fit_params.pop(fit_params.index(I_keys[0])) 
        else:
            # find the indices of the relevant fit_params and form a constraint function.
            iargs = []
            Icons = 0
            for ik,k in zip(range(len(fit_params)),fit_params):
                if k in I_keys:
                    iargs.append(ik)
                    Icons += params[k]
            cfun = lambda x: sum([x[i] for i in iargs]) - Icons 
            c_fixI0 = {'type':'eq','fun':cfun}
            c.append(c_fixI0)

    x_init = [] 
    x_bounds = [] 
    for k in fit_params:
        x_init.append(params[k])
        if k in ['r0_precursor','r0_sphere']:
            x_bounds.append((1E-3,None))
        elif k in ['I0_precursor','I0_sphere','I0_floor']:
            x_bounds.append((0.0,None))
        elif k in ['sigma_sphere']:
            x_bounds.append((0.0,0.5))

    d_opt = copy.deepcopy(params) 
    x_opt = []
    # Only proceed if there is still work to do.
    if any(x_init):
        saxs_fun = lambda q,x,d: compute_saxs_with_substitutions(q,flags,d,fit_params,x)
        I_nz = (I>0)
        n_q = len(q)
        idx_lowq = (q<0.4)
        if objfun == 'chi2':
            fit_obj = lambda x: compute_chi2( saxs_fun(q,x,params) , I )
        elif objfun == 'chi2log':
            fit_obj = lambda x: compute_chi2( np.log(saxs_fun(q[I_nz],x,params)) , np.log(I[I_nz]) )
        elif objfun == 'chi2norm':
            fit_obj = lambda x: compute_chi2( saxs_fun(q[I_nz],x,params) , I[I_nz] , weights=float(1)/I[I_nz] )
        elif objfun == 'low_q_chi2':
            fit_obj = lambda x: compute_chi2( saxs_fun(q[idx_lowq],x,params) , I[idx_lowq] )
        elif objfun == 'pearson':
            fit_obj = lambda x: -1*compute_pearson( saxs_fun(q,x,params) , I )
        elif objfun == 'low_q_pearson':
            fit_obj = lambda x: -1*compute_pearson( saxs_fun(q[idx_lowq],x,params) , I[idx_lowq] )
        elif objfun == 'low_q_chi2log':
            fit_obj = lambda x: compute_chi2( np.log(saxs_fun(q[I_nz][idx_lowq[I_nz]],x,params)) , np.log(I[I_nz][idx_lowq[I_nz]]) ) 
        elif objfun == 'pearson_log':
            fit_obj = lambda x: -1*compute_pearson( np.log(saxs_fun(q[I_nz],x,params)) , np.log(I[I_nz]) ) 
        elif objfun == 'low_q_pearson_log':
            fit_obj = lambda x: -1*compute_pearson( np.log(saxs_fun(q[I_nz][idx_lowq[I_nz]],x,params)) , np.log(I[I_nz][idx_lowq[I_nz]]) ) 
        else:
            msg = 'objective function {} not supported'.format(objfun)
            raise ValueError(msg)
        d_opt['objective_before'] = fit_obj(x_init)
        #try:
        res = scipimin(fit_obj,x_init,bounds=x_bounds,constraints=c)
        #except:
        x_opt = res.x
        d_opt['objective_after'] = fit_obj(x_opt)
        if d_opt['objective_after'] > d_opt['objective_before']:
            print('WARNING: optimization has increased the objective function. why?')
            print('x_init: {}'.format(x_init))
            print('obj_init: {}'.format(d_opt['objective_before']))
            print('x_opt: {}'.format(x_opt))
            print('obj_opt: {}'.format(d_opt['objective_after']))
            x_opt = x_init
        for k,xk in zip(fit_params,x_opt):
            d_opt[k] = xk
    return d_opt    

def compute_spherical_normal_saxs(q,r0,sigma):
    """
    Given q, a mean radius r0, 
    and the fractional standard deviation of radius sigma,
    compute the saxs spectrum assuming spherical particles 
    with normal size distribution.
    The returned intensity is normalized 
    such that I(q=0) is equal to 1.
    """
    q_zero = (q == 0)
    q_nz = np.invert(q_zero) 
    I = np.zeros(q.shape)
    if sigma < 1E-9:
        x = q*r0
        V_r0 = float(4)/3*np.pi*r0**3
        I[q_nz] = V_r0**2 * (3.*(np.sin(x[q_nz])-x[q_nz]*np.cos(x[q_nz]))*x[q_nz]**-3)**2
        I_zero = V_r0**2 
    else:
        sigma_r = sigma*r0
        dr = sigma_r*0.02
        rmin = np.max([r0-5*sigma_r,dr])
        rmax = r0+5*sigma_r
        I_zero = 0
        for ri in np.arange(rmin,rmax,dr):
            xi = q*ri
            V_ri = float(4)/3*np.pi*ri**3
            # The normal-distributed density of particles with radius r_i:
            rhoi = 1./(np.sqrt(2*np.pi)*sigma_r)*np.exp(-1*(r0-ri)**2/(2*sigma_r**2))
            I_zero += V_ri**2 * rhoi*dr
            I[q_nz] += V_ri**2 * rhoi*dr*(3.*(np.sin(xi[q_nz])-xi[q_nz]*np.cos(xi[q_nz]))*xi[q_nz]**-3)**2
    if any(q_zero):
        I[q_zero] = I_zero
    I = I/I_zero 
    return I

def precursor_heuristics(q,I,I_at_0=None):
    """
    Makes an educated guess for the radius of a small scatterer
    that would produce the input q, I(q).
    Result is bounded between 0.1 and 10 Angstroms.
    """
    # TODO: This should employ a Guinier-Porod or similar fit.
    n_q = len(q)
    # optimize the log pearson correlation in the upper half of the q domain
    nz = (I[n_q/2:]>0)
    fit_obj = lambda r: -1*compute_pearson(np.log(compute_spherical_normal_saxs(q[n_q/2:],r,0)[nz]),np.log(I[n_q/2:][nz]))
    #res = scipimin(fit_ojb,[0.1],bounds=[(0,0.3)]) 
    res = scipimin(fit_obj,[5],bounds=[(1E-3,10)])
    r_opt = res.x[0]
    return r_opt

def spherical_normal_heuristics(q,I,I_at_0=None):
    """
    This algorithm was developed and 
    originally contributed by Amanda Fournier.    

    Performs some heuristic measurements on the input spectrum,
    in order to make educated guesses 
    for the parameters of a size distribution
    (mean and standard deviation of radius)
    for a population of spherical scatterers.

    TODO: Document algorithm here.
    """
    if I_at_0 is None:
        I_at_0 = fit_I0(q,I)
    m = saxs_Iq4_metrics(q,I)
    width_metric = m['pI_qwidth']/m['q_at_Iqqqq_min1']
    intensity_metric = m['I_at_Iqqqq_min1']/I_at_0
    #######
    #
    # POLYNOMIALS FITTED FOR q0+/-10%,
    # where q0 is the argmin of a parabola
    # that is fit around the first minimum of I*q**4.
    # The function spherical_normal_heuristics_setup()
    # (in this same module) can be used to generate these polynomials.
    # polynomial for qr0 focus (with x=sigma_r/r0):
    # -8.05459639763x^2 + -0.470989868709x + 4.50108683096
    p_qr0_focus = [-8.05459639763,-0.470989868709,4.50108683096]
    # polynomial for width metric (with x=sigma_r/r0):
    # 3.12889797288x^2 + -0.0645231661487x + 0.0576604958693
    p_w = [3.12889797288,-0.0645231661487,0.0576604958693]
    # polynomial for intensity metric (with x=sigma_r/r0):
    # -1.33327411025x^3 + 0.432533640102x^2 + 0.00263776123775x + -1.27646761062e-05
    p_I = [-1.33327411025,0.432533640102,0.00263776123775,-1.27646761062e-05]
    #
    #######
    # Now find the sigma_r/r0 value that gets the extracted metrics
    # as close as possible to p_I and p_w.
    width_error = lambda x: (np.polyval(p_w,x)-width_metric)**2
    intensity_error = lambda x: (np.polyval(p_I,x)-intensity_metric)**2
    # TODO: make the objective function weight all errors equally
    heuristics_error = lambda x: width_error(x) + intensity_error(x)
    res = scipimin(heuristics_error,[0.1],bounds=[(0,0.3)]) 
    sigma_over_r = res.x[0]
    qr0_focus = np.polyval(p_qr0_focus,sigma_over_r)
    # qr0_focus = x1  ==>  r0 = x1 / q1
    r0 = qr0_focus/m['q_at_Iqqqq_min1']
    #sigma_r = sigma_over_r * r0 
    return r0,sigma_over_r

def saxs_Iq4_metrics(q,I):
    """
    From an input spectrum q and I(q),
    compute several properties of the I(q)*q^4 curve.
    This was designed for spectra that are 
    dominated by a dilute spherical form factor term.
    The metrics extracted by this Operation
    were originally intended as an intermediate step
    for estimating size distribution parameters 
    for a population of dilute spherical scatterers.

    Returns a dict of metrics.
    Dict keys and meanings:
    q_at_Iqqqq_min1: q value at first minimum of I*q^4
    I_at_Iqqqq_min1: I value at first minimum of I*q^4
    Iqqqq_min1: I*q^4 value at first minimum of I*q^4
    pIqqqq_qwidth: Focal q-width of polynomial fit to I*q^4 near first minimum of I*q^4 
    pIqqqq_Iqqqqfocus: Focal point of polynomial fit to I*q^4 near first minimum of I*q^4
    pI_qvertex: q value of vertex of polynomial fit to I(q) near first minimum of I*q^4  
    pI_Ivertex: I(q) at vertex of polynomial fit to I(q) near first minimum of I*q^4
    pI_qwidth: Focal q-width of polynomial fit to I(q) near first minimum of I*q^4
    pI_Iforcus: Focal point of polynomial fit to I(q) near first minimum of I*q^4

    TODO: document the algorithm here.
    """
    d = {}
    #if not dI:
    #    # uniform weights
    #    wt = np.ones(q.shape)   
    #else:
    #    # inverse error weights, 1/dI, 
    #    # appropriate if dI represents
    #    # Gaussian uncertainty with sigma=dI
    #    wt = 1./dI
    #######
    # Heuristics step 1: Find the first local max
    # and subsequent local minimum of I*q**4 
    Iqqqq = I*q**4
    # w is the number of adjacent points to consider 
    # when examining the I*q^4 curve for local extrema.
    # A greater value of w filters out smaller extrema.
    w = 10
    idxmax1, idxmin1 = 0,0
    stop_idx = len(q)-w-1
    test_range = iter(range(w,stop_idx))
    idx = test_range.next() 
    while any([idxmax1==0,idxmin1==0]) and idx < stop_idx-1:
        if np.argmax(Iqqqq[idx-w:idx+w+1]) == w and idxmax1 == 0:
            idxmax1 = idx
        if np.argmin(Iqqqq[idx-w:idx+w+1]) == w and idxmin1 == 0 and not idxmax1 == 0:
            idxmin1 = idx
        idx = test_range.next()
    if idxmin1 == 0 or idxmax1 == 0:
        ex_msg = str('unable to find first maximum and minimum of I*q^4 '
        + 'by scanning for local extrema with a window width of {} points'.format(w))
        d['message'] = ex_msg 
        raise Exception(ex_msg)
    #######
    # Heuristics 2: Characterize I*q**4 around idxmin1, 
    # by locally fitting a standardized polynomial.
    idx_around_min1 = (q>0.9*q[idxmin1]) & (q<1.1*q[idxmin1])
    q_min1_mean = np.mean(q[idx_around_min1])
    q_min1_std = np.std(q[idx_around_min1])
    q_min1_s = (q[idx_around_min1]-q_min1_mean)/q_min1_std
    Iqqqq_min1_mean = np.mean(Iqqqq[idx_around_min1])
    Iqqqq_min1_std = np.std(Iqqqq[idx_around_min1])
    Iqqqq_min1_s = (Iqqqq[idx_around_min1]-Iqqqq_min1_mean)/Iqqqq_min1_std
    p_min1 = np.polyfit(q_min1_s,Iqqqq_min1_s,2,None,False,np.ones(len(q_min1_s)),False)
    # polynomial vertex horizontal coord is -b/2a
    qs_at_min1 = -1*p_min1[1]/(2*p_min1[0])
    d['q_at_Iqqqq_min1'] = qs_at_min1*q_min1_std+q_min1_mean
    # polynomial vertex vertical coord is poly(-b/2a)
    Iqqqqs_at_min1 = np.polyval(p_min1,qs_at_min1)
    d['Iqqqq_min1'] = Iqqqqs_at_min1*Iqqqq_min1_std+Iqqqq_min1_mean
    d['I_at_Iqqqq_min1'] = d['Iqqqq_min1']*float(1)/(d['q_at_Iqqqq_min1']**4)
    # The focal width of the parabola is 1/a 
    p_min1_fwidth = float(1)/p_min1[0] 
    d['pIqqqq_qwidth'] = p_min1_fwidth*q_min1_std
    # The focal point is at -b/2a,poly(-b/2a)+1/(4a)
    p_min1_fpoint = Iqqqqs_at_min1+float(1)/(4*p_min1[0])
    d['pIqqqq_Iqqqqfocus'] = p_min1_fpoint*Iqqqq_min1_std+Iqqqq_min1_mean
    #######
    # Heuristics 2b: Characterize I(q) near min1 of I*q^4.
    I_min1_mean = np.mean(I[idx_around_min1])
    I_min1_std = np.std(I[idx_around_min1])
    I_min1_s = (I[idx_around_min1]-I_min1_mean)/I_min1_std
    pI_min1 = np.polyfit(q_min1_s,I_min1_s,2,None,False,np.ones(len(q_min1_s)),False)
    # polynomial vertex horizontal coord is -b/2a
    qs_vertex = -1*pI_min1[1]/(2*pI_min1[0])
    d['pI_qvertex'] = qs_vertex*q_min1_std+q_min1_mean
    # polynomial vertex vertical coord is poly(-b/2a)
    Is_vertex = np.polyval(pI_min1,qs_vertex)
    d['pI_Ivertex'] = Is_vertex*I_min1_std+I_min1_mean
    # The focal width of the parabola is 1/a 
    pI_fwidth = float(1)/pI_min1[0]
    d['pI_qwidth'] = pI_fwidth*q_min1_std
    # The focal point is at -b/2a,poly(-b/2a)+1/(4a)
    pI_fpoint = Is_vertex+float(1)/(4*pI_min1[0])
    d['pI_Ifocus'] = pI_fpoint*I_min1_std+I_min1_mean
    #######
    return d

def compute_saxs_with_substitutions(q,flags,params,x_keys,x_vals):
    p_sub = copy.deepcopy(params)
    for k,v in zip(x_keys,x_vals):
        p_sub[k] = copy.copy(v)
    return compute_saxs(q,flags,p_sub)

def compute_chi2(y1,y2,weights=None):
    """
    Compute the sum of the difference squared between input arrays y1 and y2.
    """
    if weights is None:
        return np.sum( (y1 - y2)**2 )
    else:
        weights = weights / np.sum(weights)
        return np.sum( (y1 - y2)**2*weights )

def compute_Rsquared(y1,y2):
    """
    Compute the coefficient of determination between input arrays y1 and y2.
    """
    sum_var = np.sum( (y1-np.mean(y1))**2 )
    sum_res = np.sum( (y1-y2)**2 ) 
    return float(1)-float(sum_res)/sum_var

def compute_pearson(y1,y2):
    """
    Compute the Pearson correlation coefficient between input arrays y1 and y2.
    """
    y1mean = np.mean(y1)
    y2mean = np.mean(y2)
    y1std = np.std(y1)
    y2std = np.std(y2)
    return np.sum((y1-y1mean)*(y2-y2mean))/(np.sqrt(np.sum((y1-y1mean)**2))*np.sqrt(np.sum((y2-y2mean)**2)))

def fit_I0(q,I,order=4):
    """
    Find an estimate for I(q=0) by polynomial fitting.
    All of the input q, I(q) values are used in the fitting.
    """
    I_mean = np.mean(I)
    I_std = np.std(I)
    q_mean = np.mean(q)
    q_std = np.std(q)
    I_s = (I-I_mean)/I_std
    q_s = (q-q_mean)/q_std
    p = fit_with_slope_constraint(q_s,I_s,-1*q_mean/q_std,0,order) 
    I_at_0 = np.polyval(p,-1*q_mean/q_std)*I_std+I_mean
    return I_at_0

def fit_with_slope_constraint(q,I,q_cons,dIdq_cons,order,weights=None):
    """
    Perform a polynomial fitting 
    of the low-q region of the spectrum
    with dI/dq(q=0) constrained to be zero.
    This is performed by forming a Lagrangian 
    from a quadratic cost function 
    and the Lagrange-multiplied constraint function.
    
    TODO: Explicitly document cost function, constraints, Lagrangian.

    Inputs q and I are not standardized in this function,
    so they should be standardized beforehand 
    if standardized fitting is desired.
    At the provided constraint point, q_cons, 
    the returned polynomial will have slope dIdq_cons.

    Because of the form of the Lagrangian,
    this constraint cannot be placed at exactly zero.
    This would result in indefinite matrix elements.
    """
    Ap = np.zeros( (order+1,order+1),dtype=float )
    b = np.zeros(order+1,dtype=float)
    # TODO: vectorize the construction of Ap
    for i in range(0,order):
        for j in range(0,order):
            Ap[i,j] = np.sum( q**j * q**i )
        Ap[i,order] = -1*i*q_cons**(i-1)
    for j in range(0,order):
        Ap[order,j] = j*q_cons**(j-1)
        b[j] = np.sum(I*q**j)
    b[order] = dIdq_cons
    p_fit = np.linalg.solve(Ap,b) 
    p_fit = p_fit[:-1]  # throw away Lagrange multiplier term 
    p_fit = p_fit[::-1] # reverse coefs to get np.polyfit format
    #from matplotlib import pyplot as plt
    #plt.figure(3)
    #plt.plot(q,I)
    #plt.plot(q,np.polyval(p_fit,q))
    #plt.plot(np.arange(q_cons,q[-1],q[-1]/100),np.polyval(p_fit,np.arange(q_cons,q[-1],q[-1]/100)))
    #plt.plot(q_cons,np.polyval(p_fit,q_cons),'ro')
    #plt.show()
    return p_fit
 
def spherical_normal_heuristics_setup():
    sigma_over_r = []
    width_metric = []
    intensity_metric = []
    qr0_focus = []
    # TODO: generate heuristics on a 1-d grid of sigma/r
    # instead of the 2-d grid being used here now.
    # Document proof that a 1-d grid is sufficient.
    r0_vals = np.arange(10,41,10,dtype=float)              #Angstrom
    for ir,r0 in zip(range(len(r0_vals)),r0_vals):
        q = np.arange(0.001/r0,float(10)/r0,0.001/r0)       #1/Angstrom
        sigma_r_vals = np.arange(0*r0,0.21*r0,0.01*r0)      #Angstrom
        for isig,sigma_r in zip(range(len(sigma_r_vals)),sigma_r_vals):
            I = compute_spherical_normal_saxs(q,r0,sigma_r/r0) 
            I_at_0 = compute_spherical_normal_saxs(0,r0,sigma_r/r0) 
            d = saxs_spherical_normal_heuristics(q,I)
            sigma_over_r.append(float(sigma_r)/r0)
            qr0_focus.append(d['q_at_Iqqqq_min1']*r0)
            width_metric.append(d['pI_qwidth']/d['q_at_Iqqqq_min1'])
            intensity_metric.append(d['I_at_Iqqqq_min1']/I_at_0)
    # TODO: standardize before fitting, then revert after
    p_qr0_focus = np.polyfit(sigma_over_r,qr0_focus,2,None,False,None,False)
    p_w = np.polyfit(sigma_over_r,width_metric,2,None,False,None,False)
    p_I = np.polyfit(sigma_over_r,intensity_metric,3,None,False,None,False)
    print('polynomial for qr0 focus (with x=sigma_r/r0): {}x^2 + {}x + {}'.format(p_qr0_focus[0],p_qr0_focus[1],p_qr0_focus[2]))
    print('polynomial for width metric (with x=sigma_r/r0): {}x^2 + {}x + {}'.format(p_w[0],p_w[1],p_w[2]))
    print('polynomial for intensity metric (with x=sigma_r/r0): {}x^3 + {}x^2 + {}x + {}'.format(p_I[0],p_I[1],p_I[2],p_I[3]))
    plot = True
    if plot: 
        from matplotlib import pyplot as plt
        plt.figure(1)
        plt.scatter(sigma_over_r,width_metric)
        plt.plot(sigma_over_r,np.polyval(p_w,sigma_over_r))
        plt.figure(2)
        plt.scatter(sigma_over_r,intensity_metric)
        plt.plot(sigma_over_r,np.polyval(p_I,sigma_over_r))
        plt.figure(3)
        plt.scatter(sigma_over_r,qr0_focus)
        plt.plot(sigma_over_r,np.polyval(p_qr0_focus,sigma_over_r))
        plt.figure(4)
        plt.scatter(width_metric,intensity_metric) 
        plt.show()




from collections import OrderedDict

import numpy as np
from scipy import interp
from scipy.optimize import minimize

from ... import Operation as opmod 
from ...Operation import Operation
from ....tools import saxstools

class SpectrumParameterization(Operation):
    """
    The algorithm for guessing parameters
    for the size distributions of spherical nanoparticles 
    was developed and originally contributed by Amanda Fournier.    

    This operation uses a SAXS spectrum (I(q) vs. q), 
    along with some profiling information,
    to guess a set of parameters
    for fitting a SAXS spectrum
    to theoretical scattering models. 
    TODO: document scattering models.

    A precursor is modeled by a dilute, monodisperse spherical form factor.
    TODO: Implement Guinier-Porod.
    A spherical nanoparticle population is modeled by
    a discrete sum over a probability distribution
    of dilute, monodisperse spherical form factors. 
    A crystalline arrangement is modeled by 
    a sum of pseudo-Voigt peaks.

    Outputs a return code and heuristic guesses for SAXS model parameters. 
    Also outputs the theoretical result for I(q) with the guessed parameters.

    This Operation is somewhat robust for noisy data,
    but any preprocessing (background subtraction, smoothing, or other cleaning)
    should be performed beforehand. 
    """

    def __init__(self):
        input_names = ['q', 'I', 'flags', 'fixed_params', 'fixed_param_values']
        output_names = ['params','q_I_guess']
        super(SpectrumParameterization, self).__init__(input_names, output_names)
        self.input_doc['q'] = '1d array of wave vector values in 1/Angstrom units'
        self.input_doc['I'] = '1d array of intensity values I(q)'
        self.input_doc['flags'] = 'dict of population flags- currently supported: '\
        '"bad_data", "precursor_scattering", "form_factor_scattering", "diffraction_peaks"'
        self.input_doc['fixed_params'] = 'list of strings indicating which parameters '\
        'should be fixed to specified values- these correspond one-to-one with fixed_param_values.'
        self.input_doc['fixed_param_values'] = 'list of values to which the corresponding fixed_params will be set.'

        self.output_doc['params'] = 'dict of scattering equation parameters: '\
        '"I_at_0": Intensity at q=0, by fitting the low-q region to a polynomial with dI/dq(q=0)=0; '\
        '"r0_pre": precursor term radius (Angstrom); '\
        '"I0_pre": precursor intensity scaling factor; '\
        '"r0_form": mean characteristic size (Angstrom) of form factor population; '\
        '"sigma_form": fractional standard deviation of characteristic size of form factor population; '\
        '"I0_form": form factor scattering intensity scaling factor; '\
        '"q0_structure": q value (1/Angstrom) for location of fundamental structure factor peak; '\
        '"I0_structure": structure factor intensity scaling factor; '\
        '"sigma_structure": width parameter for Pseudo-Voigt diffraction peak profile; '\
        '"R2log_guess": Coefficient of determination between logarithms of measured and parameterized spectra; '\
        '"chi2log_guess": Sum of difference squared between standardized logarithms '\
        'of measured and parameterized spectra, where both spectra are standardized '\
        'by the mean and std of the measured spectrum' 
        self.output_doc['q_I_guess'] = 'n-by-2 array of q and the intensity spectrum '\
        'corresponding to the returned scattering equation parameters'

        self.input_type['q'] = opmod.workflow_item
        self.input_type['I'] = opmod.workflow_item
        self.input_type['flags'] = opmod.workflow_item
        self.inputs['fixed_params'] = []
        self.inputs['fixed_param_values'] = []

    def run(self):
        q, I = self.inputs['q'], self.inputs['I']
        f = self.inputs['flags'] 
        if q is None or I is None or f is None:
            return
        fix_keys = self.inputs['fixed_params']
        fix_vals = self.inputs['fixed_param_values']
        p_fix = {}
        for k,v in zip(fix_keys,fix_vals):
            p_fix[k] = v

        p_new = saxstools.parameterize_spectrum(q,I,f,p_fix)

        I_guess = saxstools.compute_saxs(q,f,p_new)

        q_I_guess = np.array([q,I_guess]).T
        self.outputs['params'] = p_new
        self.outputs['q_I_guess'] = q_I_guess 







import os.path

import numpy as np
import yaml
from sklearn import preprocessing
from sklearn import linear_model

from .... import Operation as opmod 
from ....Operation import Operation
from ..... import pawstools

# this module implicitly uses sklearn.
# import here for insurance. 
import sklearn

class LoadSAXSClassifiers(Operation):
    """
    Read files to load a set of classifiers to be used on 1-d saxs spectra. 
    """

    def __init__(self):
        input_names = ['file_path']
        output_names = ['classifiers','scalers']
        super(LoadSAXSClassifiers, self).__init__(input_names, output_names)
        self.input_doc['file_path'] = 'path to a pickle file '\
        'with data defining a set of scalers and classifiers '\
        'designed for 1-d (I(q) versus q) SAXS spectra'
        self.output_doc['classifiers'] = 'a dict of sklearn '\
        'classifiers and scalers designed for 1-d SAXS spectra, '\
        'intended for input to PROCESSING.SAXS.SpectrumClassifier.'
        self.output_doc['scalers'] = 'a dict of sklearn '\
        'classifiers and scalers designed for 1-d SAXS spectra, '\
        'intended for input to PROCESSING.SAXS.SpectrumClassifier.'

        self.inputs['file_path'] = os.path.join(pawstools.sourcedir,
        'core','tools','modeling_data','scalers_and_models.yml')

    # helper function - to set parametrs for scalers and models
    def set_param(self, m_s, param):
        for k, v in param.items():
            if isinstance(v, list):
                setattr(m_s, k, np.array(v))
            else:
                setattr(m_s, k, v)

    def run(self):
        p = self.inputs['file_path']

        # load the classifiers and scalers from a file
        s_and_m_file = open(p,'rb')
        s_and_m = yaml.load(s_and_m_file)

        scalers_dict = s_and_m['scalers'] # dict of scalers parametrs
        classifier_dict = s_and_m['models'] # dict of models parametrs

        model_param_bad_data = classifier_dict['model_for_bad_data']
        model_param_form = classifier_dict['model_for_form_factor_scattering']
        model_param_precur = classifier_dict['model_for_precursor_scattering']
        model_param_diff_peaks = classifier_dict['model_for_diffraction_peaks']

        scaler_param_bad_data = scalers_dict['scaler_for_bad_data']
        scaler_param_form = scalers_dict['scaler_for_form_factor_scattering']
        scaler_param_precur = scalers_dict['scaler_for_precursor_scattering']
        scaler_param_diff_peaks = scalers_dict['scaler_for_diffraction_peaks']

        # recreate the scalers
        scaler_bad_data = preprocessing.StandardScaler()
        self.set_param( scaler_bad_data, scaler_param_bad_data)

        scaler_form = preprocessing.StandardScaler()
        self.set_param( scaler_form, scaler_param_form)

        scaler_precur = preprocessing.StandardScaler()
        self.set_param( scaler_precur, scaler_param_precur)

        scaler_diff_peaks = preprocessing.StandardScaler()
        self.set_param( scaler_diff_peaks, scaler_param_diff_peaks)

        # recreate the models
        model_bad_data = linear_model.SGDClassifier()
        self.set_param( model_bad_data, model_param_bad_data)

        model_form = linear_model.SGDClassifier()
        self.set_param( model_form, model_param_form)

        model_precur = linear_model.SGDClassifier()
        self.set_param( model_precur, model_param_precur)

        model_diff_peaks = linear_model.SGDClassifier()
        self.set_param( model_diff_peaks, model_param_diff_peaks)

        # save the dicts of classifiers and scalers as outputs
        self.outputs['classifiers'] = {'model_for_bad_data': model_bad_data,
                                       'model_for_form_factor_scattering': model_form,
                                       'model_for_precursor_scattering': model_precur,
                                       'model_for_diffraction_peaks': model_diff_peaks}
        self.outputs['scalers'] = {'scaler_for_bad_data': scaler_bad_data,
                                       'scaler_for_form_factor_scattering': scaler_form,
                                       'scaler_for_precursor_scattering': scaler_precur,
                                       'scaler_for_diffraction_peaks': scaler_diff_peaks}





from __future__ import print_function 
from PySide import QtCore

class RunnableExecutor(QtCore.QRunnable):
    """
    QRunnable that handles execution of a QWorkflow
    """

    def __init__(self,wf):
        super(RunnableExecutor,self).__init__()
        self.wf = wf

    def run(self):
        self.wf.execute()




import numpy as np
from collections import OrderedDict

from ... import Operation as opmod 
from ...Operation import Operation
from ....tools import saxstools


class SpectrumClassifier(Operation):
    """
    Classifies SAXS spectra.
    """

    def __init__(self):
        input_names = ['profiler_output', 'scalers', 'classifiers']
        output_names = ['population_flags']
        super(SpectrumClassifier, self).__init__(input_names, output_names)
        self.input_doc['profiler_output'] = 'Dict of scalar features as produced by PROCESSING.SAXS.SpectrumProfiler.'
        self.input_doc['scalers'] = 'A classification model that has been imported or trained before executing this operation.'
        self.input_doc['classifiers'] = 'A classification model that has been imported or trained before executing this operation.'
        self.output_doc['population_flags'] = 'Dict of flags indicating the presence of various scattering populations'
        self.input_type['profiler_output'] = opmod.workflow_item
        self.input_type['scalers'] = opmod.workflow_item
        self.input_type['classifiers'] = opmod.workflow_item

    def run(self):
        x = self.inputs['profiler_output']
        clsfrs = self.inputs['classifiers'] 
        sclrs = self.inputs['scalers'] 
        if x is None or clsfrs is None or sclrs is None:
            return

        # TODO: use keys: 'bad_data', 'precursor_scattering', 'form_factor_scattering', and 'diffraction_peaks',
        # i.e. same keys as population_flags output dict

        model_bad_data = clsfrs['model_for_bad_data']
        model_form = clsfrs['model_for_form_factor_scattering']
        model_precur = clsfrs['model_for_precursor_scattering']
        model_diff_peaks = clsfrs['model_for_diffraction_peaks']

        scaler_bad_data = sclrs['scaler_for_bad_data']
        scaler_form = sclrs['scaler_for_form_factor_scattering']
        scaler_precur = sclrs['scaler_for_precursor_scattering']
        scaler_diff_peaks = sclrs['scaler_for_diffraction_peaks']

        flags = OrderedDict()
        flags['test'] = True

        ### Classify the spectrum
        # (1) Apply model to input

        # transform the ordered dictionary into an array of features:
        bin_strengths = x['q_bin_strengths']
        del x['q_bin_strengths']

        #edges = x['q_bin_edges']
        del x['q_bin_edges']

        # we can use edges instead to indexes, but it could cause a problem
        # when not all samples have a specific edge
        #for i in range(len(edges)):
        #    x[str(edges[i])] = bin_strengths[i]

        # should we do it in profile_spectrum()?
        for i in range(100):
            x[str(i)] = bin_strengths[i]

        # features for bad_data, precursor, and structure labels
        features_analytical_and_60 = ['q_Imax', 'Imax_over_Imean', 'Imax_over_Ilowq',
       'Imax_over_Ihighq', 'Imax_sharpness', 'low_q_ratio', 'high_q_ratio',
       'log_fluctuation','0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13',
       '14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25',
       '26', '27', '28', '29', '30', '31', '32', '33', '34', '35', '36', '37',
       '38', '39', '40', '41', '42', '43', '44', '45', '46', '47', '48', '49',
       '50', '51', '52', '53', '54', '55', '56', '57', '58', '59' ]

        # features for form label
        features60 = ['0', '1', '2', '3', '4', '5', '6', '7', '8', '9', '10', '11', '12', '13',
       '14', '15', '16', '17', '18', '19', '20', '21', '22', '23', '24', '25',
       '26', '27', '28', '29', '30', '31', '32', '33', '34', '35', '36', '37',
       '38', '39', '40', '41', '42', '43', '44', '45', '46', '47', '48', '49',
       '50', '51', '52', '53', '54', '55', '56', '57', '58', '59']

        input_features1 = [] # analytical and 60 bins
        for item in features_analytical_and_60:
            input_features1.append(x[item])

        input_features2 = [] # 60 bins
        for item in features60:
            input_features2.append(x[item])

        transformed_input_features = scaler_bad_data.transform([input_features1])
        bad_data = model_bad_data.predict(transformed_input_features)[0] # [0] since we have only one sample
        if bad_data:
            pr_bad_data = model_bad_data.predict_proba(transformed_input_features)[0, 1]
        else:
            pr_bad_data = model_bad_data.predict_proba(transformed_input_features)[0, 0]

        flags['bad_data'] = (bad_data, pr_bad_data) # label and propability to have this label

        if bad_data == True:
            flags['precursor_scattering'] = (False, None)
            flags['form_factor_scattering'] = (False, None)
            flags['diffraction_peaks'] = (False, None)
        else:
            #form label
            transformed_input_features = scaler_form.transform([input_features2])
            form = model_form.predict(transformed_input_features)[0]
            if form:
                pr_form = model_form.predict_proba(transformed_input_features)[0, 1]
            else:
                pr_form = model_form.predict_proba(transformed_input_features)[0, 0]
            flags['form_factor_scattering'] = (form, pr_form) # label and propability to have this label

            # precursor label
            transformed_input_features = scaler_precur.transform([input_features1])
            prec = model_precur.predict(transformed_input_features)[0]
            if prec:
                pr_prec = model_precur.predict_proba(transformed_input_features)[0, 1]
            else:
                pr_prec = model_precur.predict_proba(transformed_input_features)[0, 0]
            flags['precursor_scattering'] = (prec, pr_prec)

            # difraction peaks label
            transformed_input_features = scaler_diff_peaks.transform([input_features1])
            picks = model_diff_peaks.predict(transformed_input_features)[0]
            if picks:
                pr_picks = model_diff_peaks.predict_proba(transformed_input_features)[0, 1]
            else:
                pr_picks = model_diff_peaks.predict_proba(transformed_input_features)[0, 0]
            flags['diffraction_peaks'] = (picks, pr_picks)

        # problems for later:
        # if flags['form_factor_scattering']:
        #     # classify the form factor based on inputs  
        #     flags['form_factor_id'] = ''     
        #
        # if flags['diffraction_peaks']:
        #     # classify the space group based on inputs
        #     flags['space_group_id'] = ''
        #

        self.outputs['population_flags'] = flags 





from __future__ import print_function

from .. import pawstools
from .PawsPlugin import PawsPlugin
from ..operations import Operation as opmod
        
from filestore.fs import FileStore
from metadatastore.mds import MDS
from databroker import Broker
from databroker.core import register_builtin_handlers

class DatabrokerPlugin(PawsPlugin):
    """
    Wrapper contains a Citrination client and
    implements the PawsPlugin abc interface.
    """

    def __init__(self):
        input_names = ['host','port','fs_collection','mds_collection','username','password']
        super(DatabrokerPlugin,self).__init__(input_names)
        self.inputs['fs_collection'] = 'fs_dev'
        self.inputs['mds_collection'] = 'mds_dev'

    def start(self):

        host = self.inputs['host']
        port = self.inputs['port']
        fs_collection = self.inputs['fs_collection']
        mds_collection = self.inputs['mds_collection']
        username = self.inputs['username']
        password = self.inputs['password']

        fs_config = {'host': host, 'port': port,
                     'database': fs_collection, 'username':username,'password':password}
        mds_conf = dict(database=mds_collection, host=host, username=username, password=password,
                        port=port, timezone='US/Eastern')

        fs = FileStore(fs_config, version=1)

        # may need to register handlers
        # for name,handler in handlers.iteritems(): fs.register_handler(name, handler)
        register_builtin_handlers(fs)

        mds = MDS(mds_conf, 1, auth=False)

        db = Broker(mds, fs)
        self.db = db

    def stop(self):
        pass

    @property
    def content(self): 
        return {'client':self.db,'inputs':self.inputs}

    def description(self):
        desc = str("""DataBroker API client Plugin for PAWS:
        This is a container for the PAWS Client module.
        The databroker client connects to a databroker instance
        and exposes the databroker api through self.content['client'].
        Startup requires the host address, port, and collection name;
        username and password are optionally necessary.""")
        return desc




import os
from functools import partial
from collections import OrderedDict

from PySide import QtGui, QtCore, QtUiTools

from .. import uitools
from .. import widgets
from ...core import pawstools
from ...core.operations import Operation as opmod
from ...core.models.ListModel import ListModel

class BatchWidget(QtGui.QWidget):
    """
    PAWS-based implementation of Xi-cam batch widgetry
    """

    def __init__(self,paws_api):
        super(BatchWidget,self).__init__()
        ui_file = QtCore.QFile(pawstools.sourcedir+"/ui/qtui/batch_widget.ui")
        ui_file.open(QtCore.QFile.ReadOnly)
        self.ui = QtUiTools.QUiLoader().load(ui_file)
        ui_file.close()
        self.ui.setAttribute(QtCore.Qt.WA_DeleteOnClose)
        self.paw = paws_api
        self._wfname = 'img_process'
        self._batch_wfname = 'batch'
 
        # This is a dict to translate from on-screen operation names
        # to uris for locating the Operations in paws.
        self.ops = OrderedDict()
        self.ops['read_PONI'] = 'IO.CALIBRATION.ReadPONI'
        #self.ops['read_darkfield'] = 'INPUT.TIF.LoadTif'
        #self.ops['read_flatfield'] = 'INPUT.TIF.LoadTif'
        #self.ops['read_background'] = 'INPUT.TIF.LoadTif'
        self.ops['read_image'] = 'IO.IMAGE.FabIOOpen'
        self.ops['calibrate_image'] = 'PROCESSING.CALIBRATION.Calibrate'
        self.ops['integrate_image'] = 'PROCESSING.CALIBRATION.CalReduce'
        self.ops['log_I_2d'] = 'PROCESSING.BASIC.ArrayLog'
        self.ops['log_I_1d'] = 'PROCESSING.BASIC.LogY'
        self.ops['write_csv'] = 'IO.CSV.WriteArrayCSV'
        #self.ops['batch'] = 'EXECUTION.BATCH.BatchFromFiles'

        self.paw.activate_op('EXECUTION.BATCH.BatchFromFiles')
        for nm,opuri in self.ops.items():
            self.paw.activate_op(opuri)       

        # This is a dict for holding on to widgets,
        # so that they can be displayed in self.ui.output_tabs as needed.
        self.output_widgets = {} 

        self.build_ui()
        self.wf_setup()

    def build_ui(self):
        self.ui.output_tabs.clear()
        self.ui.setWindowTitle('batch widget')
        self.ui.wf_box.setTitle('workflow control')
        self.ui.batch_box.setTitle('batch control')
        self.ui.fs_box.setTitle('filesystem browser')
        self.ui.viewer_box.setTitle('viewer')
        self.ui.run_wf_button.setText('&Run') 
        self.ui.run_wf_button.clicked.connect(self.run_wf) 
        self.ui.setStyleSheet( "QLineEdit { border: none }" + self.ui.styleSheet() )
        self.ui.main_splitter.setSizes([1,2,1])
        self.ui.add_files_button.setText('Add selected files')
        self.ui.add_files_button.clicked.connect(self.add_files)
        self.ui.remove_files_button.setText('Remove selected files')
        self.ui.remove_files_button.clicked.connect(self.rm_files)
        # filesystem viewer setup
        fsmod = QtGui.QFileSystemModel()
        fsmod.setRootPath(QtCore.QDir.currentPath())
        self.ui.fs_browser.setModel(fsmod)
        self.ui.fs_browser.hideColumn(1)
        self.ui.fs_browser.hideColumn(3)
        self.ui.fs_browser.setColumnWidth(0,400)
        idx = fsmod.index(QtCore.QDir.currentPath())
        while idx.isValid():
            self.ui.fs_browser.setExpanded(idx,True)
            idx = idx.parent()
        #self.ui.fs_browser.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        #self.ui.fs_browser.setSelectionModel(QtGui.QItemSelectionModel.Rows)
        self.ui.batch_list.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)
        self.ui.fs_browser.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)

    def add_files(self):
        idxs = self.ui.fs_browser.selectedIndexes()
        for idx in idxs:
            if idx.column() == 0:
                fpath = self.ui.fs_browser.model().filePath(idx)
                itm = QtGui.QListWidgetItem()
                itm.setText(fpath)
                self.ui.batch_list.addItem(itm)    

    def rm_files(self):
        itms = self.ui.batch_list.selectedItems()
        for itm in itms:
            self.ui.batch_list.takeItem(self.ui.batch_list.row(itm))

    def wf_setup(self):
        self.paw.add_wf(self._wfname)
        self.paw.connect_wf_input('image_path','read_image.inputs.path',self._wfname)
        self.paw.add_wf(self._batch_wfname)

        # Set up the batch execution Operation first
        self.paw.select_wf(self._batch_wfname)
        self.paw.add_op('batch','EXECUTION.BATCH.BatchFromFiles')
        self.paw.set_input('batch','workflow',self._wfname)
        self.paw.set_input('batch','input_name','image_path')

        # Set up the rest of the workflow
        self.paw.select_wf(self._wfname)
        for op_name,op_uri in self.ops.items():
            add_op_row = self.paw.op_count(self._wfname)+1
            op_tag = op_name

            # Add the op to the workflow
            self.paw.add_op(op_tag,op_uri,self._wfname)
            # Set up the inputs....
            self.default_op_setup(op_tag)

            # Add the op name in a text box
            op_tag_widget = QtGui.QLineEdit(op_tag)
            op_tag_widget.setReadOnly(True)
            self.ui.wf_layout.addWidget(op_tag_widget,add_op_row,0,1,1)

            # Add buttons to interact with op 
            op_edit_button = QtGui.QPushButton('edit')
            op_edit_button.clicked.connect( partial(self.edit_op,op_tag,op_name) )
            enable_toggle = QtGui.QCheckBox('enable')
            enable_toggle.setCheckState(QtCore.Qt.Checked)
            #enable_toggle.stateChanged.connect( partial(self.toggle_enabled,op_tag) )
            vis_toggle = QtGui.QCheckBox('view')
            vis_toggle.stateChanged.connect( partial(self.set_visualizer,op_tag) )
            self.ui.wf_layout.addWidget(enable_toggle,add_op_row,3,1,1)
            self.ui.wf_layout.addWidget(vis_toggle,add_op_row,4,1,1)
            self.ui.wf_layout.addWidget(op_edit_button,add_op_row,5,1,1)

    def default_op_setup(self,op_tag):
        if op_tag == 'read_PONI':
            # Ultimately this should be set by the user or taken from another part of the application
            #self.paw.set_input(op_tag,'poni_file','')
            self.paw.set_input(op_tag,'poni_file',os.path.join(pawstools.paws_scratch_dir,'test.poni'))

        elif op_tag == 'read_image':
            # This is where the batch operation will set its inputs
            self.paw.set_input(op_tag,'path','')

        elif op_tag == 'calibrate_image' or op_tag == 'integrate_image':
            self.paw.set_input(op_tag,'image_data','read_image.outputs.image_data')
            self.paw.set_input(op_tag,'poni_dict','read_PONI.outputs.poni_dict')

        elif op_tag == 'log_I_1d':
            self.paw.set_input(op_tag,'x_y','integrate_image.outputs.q_I')

        elif op_tag == 'log_I_2d':
            self.paw.set_input(op_tag,'x','calibrate_image.outputs.I_at_q_chi')

        elif op_tag == 'write_csv':
            self.paw.set_input(op_tag,'array','integrate_image.outputs.q_I')
            self.paw.set_input(op_tag,'headers',['q','I'])
            self.paw.set_input(op_tag,'dir_path','read_image.outputs.dir_path','workflow item')
            self.paw.set_input(op_tag,'filename','read_image.outputs.filename')
            self.paw.set_input(op_tag,'filetag','_processed')

    def set_visualizer(self,op_tag,state):
        if not state==0:
            # Find, create, or otherwise open the widget
            if not op_tag in self.output_widgets.keys():
                widg = self.make_widget(op_tag) 
                self.output_widgets[op_tag] = widg
            else:
                # The user closed the tab
                # instead of un-checking the visualizer box,
                # so the widget should still be in self.output_widgets
                widg = self.output_widgets[op_tag]

            if self.ui.output_tabs.indexOf(widg) == -1:
                # Add a tab to self.ui.output_tabs 
                tab_idx = self.ui.output_tabs.addTab(widg,op_tag) 
            self.ui.output_tabs.setCurrentWidget(widg)
        else:
            widg = self.output_widgets.pop(op_tag)
            if widg is not None:
                tab_idx = self.ui.output_tabs.indexOf(widg)  
                widg.close() 
                if not tab_idx == -1: 
                    self.ui.output_tabs.removeTab(tab_idx)

    def make_widget(self,op_tag):
        if op_tag == 'read_PONI':
            output_data = self.paw.get_output(op_tag,'poni_dict',self._wfname)
        elif op_tag == 'read_image':
            output_data = self.paw.get_output(op_tag,'image_data',self._wfname)
        elif op_tag == 'calibrate_image':
            output_data = self.paw.get_output(op_tag,'I_at_q_chi',self._wfname)
        elif op_tag == 'integrate_image':
            output_data = self.paw.get_output(op_tag,'q_I',self._wfname)
        elif op_tag == 'log_I_1d':
            output_data = self.paw.get_output(op_tag,'x_logy',self._wfname)
        elif op_tag == 'log_I_2d':
            output_data = self.paw.get_output(op_tag,'logx',self._wfname)
        elif op_tag == 'write_csv':
            output_data = self.paw.get_output(op_tag,'csv_path',self._wfname)
        # Form a widget from the output data 
        widg = widgets.make_widget(output_data)
        return widg

    def run_wf(self):
        self.paw.select_wf(self._batch_wfname)
        file_list = []
        nfiles = self.ui.batch_list.count()
        for r in range(nfiles):
            p = self.ui.batch_list.item(r).text()
            file_list.append(p)
        self.paw.set_input('batch','file_list',file_list)
        self.paw.execute()
        self.update_visuals()

    def update_visuals(self):
        for widg in self.output_widgets:
            if isinstance(widg,QtGui.QWidget):
                widg.repaint()

    def edit_op(self,op_tag,op_name):
        pass




import numpy as np

from ... import Operation as opmod 
from ...Operation import Operation
from ... import optools

class BuildListFromBatch(Operation):
    """
    Given a batch output and a batch output uri, 
    harvest a list of outputs from the batch.
    """

    def __init__(self):
        input_names = ['batch_output','output_key']
        output_names = ['data_list']
        super(BuildListFromBatch,self).__init__(input_names,output_names)        
        self.input_doc['batch_output'] = 'list of dicts produced by a batch execution'
        self.input_doc['output_key'] = 'dict key for batch_output data to be packed into list'
        self.input_type['batch_output'] = opmod.workflow_item
        self.output_doc['data_list'] = 'list of the data fetched from batch_output using data_uri'

    def run(self):
        b_out = self.inputs['batch_output']
        k = self.inputs['output_key']
        if b_out is None or k is None:
            return
        l = [d[k] for d in b_out]
        self.outputs['data_list'] = l 

